/**
 * Introduction
 *
 * @title ROV - Easy applications from C
 *
 * @copyright Jason Tribbeck / 7th ARM 1994-2001
 *
 * @authors Jason Tribbeck
 *
 * @file drawrender.c
 */

#include <string.h>

#include "main.h"
#include "draw.h"

#define scx(xv) (ox+((int)((double)(xv)*ssx)))
#define scy(yv) (oy+((int)((double)(yv)*ssy)))

#define chkbox(qq) ((scx(file[qq]-zx)>x1)||(scy(file[qq+1]-zy)>y1)||(scx(file[qq+2]-zx)<x0)||(scy(file[qq+3]-zy)<y0))

void draw_setfont(int handle);
void draw_selecttable(int, int *, int, int, char *);

extern int __draw_localx, __draw_localy;

/* #define __draw_localx (0)
   #define __draw_localy (0)*/

extern int draw_wimppalette[], print_inprint;

static int ox, oy, x0, y0, x1, y1;
static int * eof, zx, zy;
static double scale_x, scale_y, ssx, ssy;

static int draw_trans[6];
static char *font_trans[256];

static void render(int *, int *);

static void resolve(int *file, int * xa, int * ya, int * xb, int * yb)
{
  *xa = scx(file[2] - zx);
  *ya = scy(file[3] - zy);
  *xb = scx(file[4] - zx);
  *yb = scy(file[5] - zy);
}

static void render_obj(int * file)
{
  int *sof = file + 6, sod = 4;

  switch(file[0]) {
  case 6 :
    render(&file[3 + 6], & file[file[1] >> 2]);
    break;
  case 2 :
    if(sof[3] & 128)
      sod = 6 + sof[5];
    if(sof[0] != -1) {
      draw_setcolour(sof[0]);
      draw_fill(sof + sod, ((sof[3] >> 5) & 2) | 48, draw_trans, 0);
    }
    if(sof[1] != -1) {
      draw_capline l;

      l.join = sof[3] & 3;
      l.lead_cap = (sof[3] >> 2) & 3;
      l.tail_cap = (sof[3] >> 4) & 3;
      l.reserved = 0;
      l.mitre_limit = 10 << 16;
      l.lead_capwidth = l.tail_capwidth = ((sof[3] >> 16) & 255) << 4;
      l.lead_caplength = l.tail_caplength = ((sof[3] >> 24) & 255) << 4;
      draw_setcolour(sof[1]);
      if(sof[3] & 128)
	draw_stroke(sof + sod, 0, draw_trans, 0, sof[2], & l, (draw_dash *)sof + 4);
      else
	draw_stroke(sof + sod, 0, draw_trans, 0, sof[2], & l, 0);
    }
    break;
  case 1 :
    if(sof[2]) {
      int fonth;
      int xa, ya;
      fonth = draw_findfont(font_trans[sof[2]], (int)(((double)(sof[3])) * scale_x / 40.0), \
			    (int)(((double)(sof[4])) * scale_y / 40.0));
      if(fonth == -1)
	return;
      xa = scx(sof[5] - zx) + __draw_localx;
      ya = scy(sof[6] - zy) + __draw_localy;
      draw_setfontcolour(fonth, sof[1], sof[0]);
      draw_setfont(fonth);
      draw_paintfont(fonth, (char *)(sof + 7), xa, ya);
      draw_losefont(fonth);
    } else {
      int xa, ya, xb, yb, trans[4], i, ns;
      unsigned char * c = (unsigned char *) & sof[7];

      xa = scx(file[2] - zx);
      ya = scy(file[3] - zy);
      xb = scx(file[4] - zx);
      yb = scy(file[5] - zy);
      ns = ((int)((double)sof[3] * scale_x));
      resolve(file, & xa, & ya, & xb, & yb);
      draw_setcolour(sof[0]);

      trans[0] = (xb - xa) * 2 / draw_scalex;
      trans[1] = (yb - ya) * 4 / draw_scaley;
      trans[2] = strlen((char*)c) * 16;
      trans[3] = 32;
      if(print_inprint == 0)
	for(i = 0; c[i]; i ++)
	  draw_charscaled(c[i], xa + (i * ns) / 280 + __draw_localx, ya + __draw_localy, trans);
      else {
	draw_setcharsize(ns / 280, trans[1] / 8);
	for(i = 0; c[i]; i ++) {
	  draw_move(xa + (i * ns) / 280 + __draw_localx, yb + __draw_localy);
	  draw_pchar(c[i]);
	}
      }
    }
    break;
  case 12 :
    if(sof[9]) {
      int fonth;
      int xa, ya, trans[6];
      fonth = draw_findfont(font_trans[sof[9]], (int)(((double)(sof[10])) / 40.0), \
			    (int)(((double)(sof[11])) / 40.0));
      if(fonth == -1)
	return;
      trans[0] = (int)((double)sof[0] * scale_x);
      trans[1] = (int)((double)sof[1] * scale_y);
      trans[2] = (int)((double)sof[2] * scale_x);
      trans[3] = (int)((double)sof[3] * scale_y);
      trans[4] = sof[4];
      trans[5] = sof[5];
      /* trans[4] = (int)((double)(trans[4] - zx) * scale_x);
	 trans[5] = (int)((double)(trans[5] - zy) * scale_y); */
      xa = scx(sof[12] - zx);
      ya = scy(sof[13] - zy);
      draw_setfontcolour(fonth, sof[8], sof[7]);
      draw_setfont(fonth);
      draw_paintfontx(fonth, (char *)(sof + 14), (xa + __draw_localx) * 400, (ya + __draw_localy) * 400, trans, sof[6] << 9);
      draw_losefont(fonth);
    }
    break;
  case 5 :
    {
      int xa, ya, xb, yb, trans[4], px, py, mode, cols;
      char ctrans[256];

      resolve(file, & xa, & ya, & xb, & yb);
      draw_getspriteinfo(sof, & px, & py, & mode);
      /* px = px * draw_scalex;
	 py = py * draw_scaley;
	 py = py << (draw_getmodevar(mode, 5)); */
      trans[0] = xb - xa;
      trans[1] = yb - ya;
      trans[2] = px * draw_scalex;
      trans[3] = py * draw_scaley;
      cols = draw_getmodevar(mode, 3);
      if((sof[8] != 44) && (sof[8] != 0x22c)) {
	int table[256], i, ic;

	if((ic = cols) == 63) ic = 255;
	for(i = 0; i <= ic; i ++)
	  table[i] = sof[12 + i * 2];
	draw_selecttable(mode, table, - 1, - 1, ctrans);
      }   else {
	if(cols < 63)
	  draw_selecttable(mode, draw_wimppalette, - 1, - 1, ctrans);
	else
	  draw_selecttable(mode, 0, - 1, - 1, ctrans);
      }
      draw_plotsprite((void *)sof, 8, xa + __draw_localx, ya + __draw_localy, trans, ctrans);
    }
    break;
  case 13 :
    {
      int trans[6], px, py, mode, cols;
      char ctrans[256];

      draw_getspriteinfo(sof + 6, & px, & py, & mode);
      cols = draw_getmodevar(mode, 3);
      if((sof[8 + 6] != 44) && (sof[8 + 6] != 0x22c)) {
	int table[256], i, ic;

	if((ic = cols) == 63) ic = 255;
	for(i = 0; i <= ic; i ++)
	  table[i] = sof[12 + 6 + i * 2];
	draw_selecttable(mode, table, - 1, - 1, ctrans);
      } else {
	if(cols < 63)
	  draw_selecttable(mode, draw_wimppalette, - 1, - 1, ctrans);
	else
	  draw_selecttable(mode, 0, - 1, - 1, ctrans);
      }
      trans[0] = (int)(((double)sof[0] * scale_x));
      trans[1] = (int)(((double)sof[1] * scale_y));
      trans[2] = (int)(((double)sof[2] * scale_x));
      trans[3] = (int)(((double)sof[3] * scale_y));
      trans[4] = (int)(((double)(sof[4] - zx)) * scale_x) + ((__draw_localx + ox) << 8);
      trans[5] = (int)(((double)(sof[5] - zy)) * scale_y) + ((__draw_localy + oy) << 8);
      draw_plotsprite56((void *)(sof + 6), 8, trans, ctrans);
    }
    break;
  default :
    /* {
       int xa, ya, xb, yb;
       resolve(file, & xa, & ya, & xb, & yb);
       draw_move(xa, ya);
       draw_line(xb, ya);
       draw_line(xb, yb);
       draw_line(xa, yb);
       draw_line(xa, ya);
       draw_move((xa + xb) / 2, (ya + yb) / 2);
       printf("%d", file[0]);
       } */
    break;
  }
}

static void make_fonts(char * file, char * efile)
{
  while((int)file < (int)efile) {
    font_trans[(int)(* file)] = file + 1;
    file ++ ;
    while(* file ++);
    /* file = (char *)((((int)file) + 3) & 0xfffffffc); */
  }
}

static void render(int * file, int * efile)
{
  while((int)file < (int)efile) {
    if((file[0] != 0) && (file[0] != 0x0b)) {
      if chkbox(2) {
      } else {
	render_obj(file);
      }
    }
    if(file[0] == 0)
      make_fonts((char *)(file + 2), (char *)(& file[file[1] >> 2]));
    file = & file[file[1] >> 2];
  }
}

int drawrender_xsize(int * file, double sx)
{
  if((file[0] != 0x77617244) && (file[1] > 201)) return 0;
  return (int)((double)(file[8] - file[6]) * sx / 256.0);
}

int drawrender_ysize(int * file, double sx)
{
  if((file[0] != 0x77617244) && (file[1] > 201)) return 0;
  return (int)((double)(file[9] - file[7]) * sx / 256.0);
}

void drawrender_render(int * file, int fsize, int zox, int zoy, int zx0, int zy0, int zx1, int zy1, double sx, double sy)
{
  memset(font_trans, 0, 1024);
  scale_x = sx;
  scale_y = sy;
  ssx = scale_x / 256.0;
  ssy = scale_y / 256.0;
  draw_trans[0] = (int)(65536.0 * sx);
  draw_trans[3] = (int)(65536.0 * sy);
  draw_trans[1] = draw_trans[2] = 0;
  ox = zox; oy = zoy; x0 = zx0; y0 = zy0; x1 = zx1; y1 = zy1;
  eof = & file[fsize >> 2];
  if(file[0] != 0x77617244)
    return;
  if(file[1] > 201)
    return;
  file += 6;
  zx = file[0];
  zy = file[1];
  draw_trans[4] = (ox + __draw_localx) * 256 - ((int)((double)zx * sx));
  draw_trans[5] = (oy + __draw_localy) * 256 - ((int)((double)zy * sy));
  if chkbox(0) return;
  file += 4;
  render(file, eof);
}
